using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Text.RegularExpressions;

namespace FFXISwitchMon {
   public partial class fMonitor : Form {
      //Constants
      private const string classmatch = "FFXiClass";
      private const string polname = "xiloader.exe";
      private const string config_key = "SOFTWARE\\FFXISwitchMon";
      public  const string appversion = "1.1.0x";

      //Application Pieces
      private List<FFXIWindow> FFXIWindows;
      private Dictionary<int, ActionCommand> bindlist;

      private Joy.Joysticks joys;
      private Joy.Joystick  joystick;
      private Options config;      
      private fOptions opt;
      private MagSets MagSetManager;
      private Timer tGather;

      //Local config
      private int currentindex = 0;
      private int m_gather_mode;

      //Game Window Events
      public event windowevent  WindowAdded;
      public event windowevent  WindowRemoved;
      public event windowevent  WindowChanged;
      public event genericevent OptionsUpdated;   
           
      //==========================================================================================
      // Constructor
      //==========================================================================================
      public fMonitor() {
         joys = new Joy.Joysticks();
         InitializeComponent();

         //Load the configuration
         Dictionary<string, object> defaults = new Dictionary<string, object>();
         defaults["gathermode"] = 2;

         config = new Options(config_key, defaults);
         bindlist = new Dictionary<int,ActionCommand>();
         bindlist[config.GetValue<int>("buttonflipnext" , -1)] = ActionCommand.ActivateNext;
         bindlist[config.GetValue<int>("buttonflipprev" , -1)] = ActionCommand.ActivatePrev;
         bindlist[config.GetValue<int>("buttonmaxtoggle", -1)] = ActionCommand.ToggleZoom;
         bindlist[config.GetValue<int>("buttonmagtoggle", -1)] = ActionCommand.ToggleMagnifiers;
         bindlist[config.GetValue<int>("buttoncycleset" , -1)] = ActionCommand.CycleSets;
         bindlist[config.GetValue<int>("buttonopenmenu" ,  5)] = ActionCommand.OpenMenu;              

         m_gather_mode = config.GetValue<int>("gathermode", 2);

         //Detect FFXI game windows
         FFXIWindows = new List<FFXIWindow>();
         EnumFFXIWindows();

         //Initialize Joysticks
         if (joys.Count > 0) {
            //Load the previous joystick
            if (config["lastjoy"] != null) {
               Guid joyguid = new Guid((string)config["lastjoy"]);
               joystick = joys.FindJoystick(joyguid);
            }
            //If no previous joystick the select the first one
            if (joystick == null) {
               joystick = joys[0];
               config["lastjoy"] = joystick.ID;
               config.Save();
            }
            //Begin polling joystick
            joystick.ButtonPressDown += new Joy.ButtonPressCallback(joystick_ButtonPressDown);
            joystick.Acquire();
         } else {
            joystick = null;
         }

         //Initialize Magnifiers
         MagSetManager = new MagSets(this, this.config);
         foreach (KeyValuePair<string, MagSet> pair in MagSetManager) {
            ToolStripItem newitem;
            newitem = this.miChangeMagSet.DropDownItems.Add(pair.Value.Name);
            newitem.Click += new EventHandler(miChangeMagSet_SubItem_Click);
            newitem.Tag = pair.Value;

            newitem = this.miRemoveMagSet.DropDownItems.Add(pair.Value.Name);
            newitem.Click += new EventHandler(miRemoveMagSet_SubItem_Click);
            newitem.Tag = pair.Value;
         }
         MagSetManager_SetChanged(MagSetManager.ActiveSet);
         MagSetManager.SetCreated += new MagnifierSetCallback(MagSetManager_SetCreated);
         MagSetManager.SetRemoved += new MagnifierSetCallback(MagSetManager_SetRemoved);
         MagSetManager.SetChanged += new MagnifierSetCallback(MagSetManager_SetChanged);
         
         tGather = new Timer();
         tGather.Tick += new EventHandler(tGather_Tick);
         tGather.Interval = config.GetValue<int>("gatherinterval", 1000);
         tGather.Enabled = (m_gather_mode == 1);

         //Setup Events
         this.OptionsUpdated += new genericevent(fMonitor_OptionsUpdated);
      }

      //==========================================================================================
      // Properties
      //==========================================================================================
      public Joy.Joysticks Joysticks {
         get { return this.joys; }
      }
      public Options Config {
         get { return this.config; }
      }
      public List<FFXIWindow> WindowList {
         get { return this.FFXIWindows; }
      }
      public Joy.Joystick ActiveJoystick {
         get { return joystick; }
         set {
            if (value != joystick) {
               joystick.Unacquire();
               joystick = value;
               joystick.Acquire();
            }
         }
      }
      public Dictionary<int, ActionCommand> BindList {
         get { return bindlist; }        
      }
      public MagSets MagManager {
         get { return this.MagSetManager; }
      }
      
      //==========================================================================================
      // Utility Functions
      //==========================================================================================
      /// <summary>
      /// Enumerate the current game windows loaded on your system
      /// </summary>
      private void EnumFFXIWindows() {
         EnumFFXIWindows(false);
      }
      private void EnumFFXIWindows(bool NoActivateNext) {
         FFXIWindow current = this.CurrentWindow;
         int idx = this.currentindex;

         //Mark all existing windows as "Dirty"
         foreach (FFXIWindow window in this.FFXIWindows) {
            window.dirty = true;
#if FAKEWND && DEBUG
            if (window.classname == "FAKEWND")
               window.dirty = false;
#endif
         }

         WinAPI.EnumWindows(new EnumWindowCallBack(EnumWindowCallBack), 0);

         //Remove any window that remained dirty through enumeration
         for (int index = this.FFXIWindows.Count - 1; index >= 0; --index) {
            FFXIWindow window = this.FFXIWindows[index];
            if (window.dirty) {
               this.FFXIWindows.Remove(window);
               if (this.WindowRemoved != null)
                  this.WindowRemoved(window);
            }
         }

#if FAKEWND && DEBUG
         //Add Fake Windows
         if (this.FFXIWindows.Count <= 0) {
            IntPtr hWnd = WinAPI.GetDesktopWindow();
            //Add the window to the internal list
            for (int i = 0; i < 3; i++) {
               FFXIWindow item = new FFXIWindow();
               item.hwnd = hWnd;
               item.classname = "FAKEWND";
               this.FFXIWindows.Add(item);
            }         
         }
#endif

         //Fix current index or update magnifier bindings if game window was killed
         if (this.FFXIWindows.Count > 0) {
            int newidx = this.FFXIWindows.IndexOf(current);
            if (newidx >= 0 && (idx != newidx)) {
               this.currentindex = newidx;
            } else if (newidx < 0 && !NoActivateNext) {
               ActiveNextWindow();
            }
         }
      }

      /// <summary>
      /// Game windows change title text every couple seconds and also have a scrambled class name making
      /// conventional detection harder. However, Windows will still report the executable name that owns
      /// the window handle when using the correct API calls and the class name can never change for the
      /// life of the window.
      /// 
      /// Since obtaining the module name for hundreds of windows can be time consuming, at least in the
      /// scope of the .NET framework, a few tricks are employed to reduce the number of windows to check:
      /// 
      /// First,  we are only interested in windows that have no parent which will reduce the
      ///         canidates substantially, depending on what other applications are running at the time.
      /// Second, we compare the class name against the base ffxi window class
      ///
      /// Updated (11.29.2007) Douglas (nekura)
      ///  FFXI no longer randomly generates a class name, instead it always uses FFXiClass.
      ///  This was changed in the update that introduced windowed mode.
      /// 
      /// For archival and historical purposes the old detection logic may still be referenced
      /// within svn commits 1.01 and below, in the event it is added again in a future patch.
      /// 
      /// </summary>
      /// <param name="hWnd">Handle of the window</param>
      /// <param name="lParam">Unused</param>
      /// <returns>Boolean response back to the API which generated this event</returns>
      private bool EnumWindowCallBack(IntPtr hWnd, int lParam) {
         try {
            //Only look at top-level windows
            if (WinAPI.GetParent(hWnd) <= 0) {
               //Retrieve the window ClassName
               StringBuilder classname = new StringBuilder(256);
               WinAPI.GetClassName(hWnd, classname, 256);

               //Looking up the module name has a time hit, so reduce the amount of canidates by using a regular expression match
               if (classmatch.Equals(classname.ToString())) {
                  //Retrieve the Process ID
                  uint pid;
                  WinAPI.GetWindowThreadProcessId(hWnd, out pid);

                  //Open the Process and begin probe
                  IntPtr hProcess = WinAPI.OpenProcess(WinAPI.PROCESS_VM_READ | WinAPI.PROCESS_QUERY_INFORMATION, 0, pid);
                  if (hProcess.ToInt32() > 0) {
                     //Only retrieve the first module for the process since it is the executable
                     uint[] modlist = new uint[1];
                     uint count;
                     if (WinAPI.EnumProcessModules(hProcess, modlist, sizeof(uint), out count)) {
                        //Retrieve the executable's file name
                        StringBuilder module = new StringBuilder(256);
                        if (WinAPI.GetModuleBaseName(hProcess, (IntPtr)modlist[0], module, 256) > 0) {
                           if ((module.ToString() == polname) || (module.ToString() == "pol.exe")) {
                              //Search for the window in the current collection. If already exists mark it and exit
                              foreach (FFXIWindow window in this.FFXIWindows) {
                                 if (window.hwnd == hWnd) {
                                    window.dirty = false;
                                    WinAPI.CloseHandle(hProcess);
                                    return true;
                                 }
                              }
                           
                              //Add the window to the internal list
                              FFXIWindow item = new FFXIWindow();
                              item.hwnd = hWnd;
                              item.classname = classname.ToString();
                              this.FFXIWindows.Add(item);

                              if (this.WindowAdded != null)
                                 this.WindowAdded(item);
                           }
                        }
                     }
                     WinAPI.CloseHandle(hProcess);
                  }
               }
            }
         } catch (Exception ex) {
#if DEBUG
            System.Diagnostics.Trace.WriteLine("EnumWindowCallBack Error: " + ex.Message);
#endif
         }
         return true;
      }

      private bool WindowExists(List<FFXIWindow> array, FFXIWindow item) {
         foreach (FFXIWindow candidate in array) {
            if (candidate.hwnd == item.hwnd)
               return true;
         }
         return false;
      }

      /// <summary>
      /// Windows 2000 and XP will refuse to activate and bring windows to the foreground
      /// when the calling thread is not the foreground thread itself.
      /// 
      /// As a work around, This function will attach itself to whichever application is the
      /// foreground process and essentially trick windows into doing our bidding. (evil laugh).
      /// </summary>
      /// <param name="hWnd">Handle of the target window to activate and switch focus to</param>
      private void ForceForegroundWindow(IntPtr hWnd) {
         uint a;
    
         IntPtr hWndForeground = WinAPI.GetForegroundWindow();
         if (hWndForeground != hWnd) {
            uint thread1 = WinAPI.GetWindowThreadProcessId(hWndForeground, out a);
            uint thread2 = WinAPI.GetCurrentThreadId();

            if (thread1 != thread2) {
               WinAPI.AttachThreadInput(thread1, thread2, true);
               WinAPI.BringWindowToTop(hWnd);
               if (WinAPI.IsIconic(hWnd)) {
                  WinAPI.ShowWindow(hWnd, WinAPI.ShowWindowFlags.SW_SHOWNORMAL);
               } else {
                  WinAPI.ShowWindow(hWnd, WinAPI.ShowWindowFlags.SW_SHOW);
               }
               WinAPI.AttachThreadInput(thread1, thread2, false);
            } else {
               WinAPI.SetForegroundWindow(hWnd);
            }
            if (WinAPI.IsIconic(hWnd)) {
               WinAPI.ShowWindow(hWnd, WinAPI.ShowWindowFlags.SW_SHOWNORMAL);
            } else {
               WinAPI.ShowWindow(hWnd, WinAPI.ShowWindowFlags.SW_SHOW);
            }
         }
      }

      public FFXIWindow CurrentWindow {
         get {
            if (this.FFXIWindows.Count > 0 && currentindex < this.FFXIWindows.Count)
               return this.FFXIWindows[currentindex];
            return null;
         }
      }

      /// <summary>
      /// Flips to the next game window on the stack
      /// </summary>
      internal void ActiveNextWindow() {
#if DEBUG
         System.Diagnostics.Trace.WriteLine("ACTION: Activate Next");
#endif
         if (m_gather_mode == 2) {
            EnumFFXIWindows(true);
         }

         int attemptcount = 0;
         int attemptidx = currentindex;
         if (this.FFXIWindows.Count > 0) {
            //prevent run-over when all active windows are denied flip access
            while (attemptcount <= this.FFXIWindows.Count) {
               attemptidx += 1;
               if (attemptidx >= this.FFXIWindows.Count)
                  attemptidx = 0;

               FFXIWindow item = this.FFXIWindows[attemptidx];
               if (!item.unallowactivate) {
                  currentindex = attemptidx;
                  ForceForegroundWindow(item.hwnd);
                  if (WindowChanged != null)
                     WindowChanged(item);
                  return;
               }               
               attemptcount += 1;
            }
         }
      }

      /// <summary>
      /// Flips to the previous game window on the stack
      /// </summary>
      internal void ActivePreviousWindow() {
#if DEBUG
         System.Diagnostics.Trace.WriteLine("ACTION: Activate Previous");
#endif
         int attemptcount = 0;
         int attemptidx = currentindex;

         if (m_gather_mode == 2) {
            EnumFFXIWindows(true);
         }
      
         if (this.FFXIWindows.Count > 0) {
            //prevent run-over when all active windows are denied flip access
            while (attemptcount <= this.FFXIWindows.Count) {
               attemptidx -= 1;
               if (attemptidx < 0)
                  attemptidx = this.FFXIWindows.Count - 1;

               FFXIWindow item = this.FFXIWindows[attemptidx];
               if (!item.unallowactivate) {
                  currentindex = attemptidx;
                  ForceForegroundWindow(item.hwnd);
                  if (WindowChanged != null)
                     WindowChanged(item);
                  return;
               }
               attemptcount += 1;
            }
         }
      }
      
      /// <summary>
      /// Activate the current window
      /// </summary>
      internal void ActivateCurrentWindow() {
#if DEBUG
         System.Diagnostics.Trace.WriteLine("ACTION: Activate Current");
#endif
         if (m_gather_mode == 2) {
            EnumFFXIWindows(true);
         }

         if (this.FFXIWindows.Count > 0 && currentindex < this.FFXIWindows.Count) {
            FFXIWindow item = this.FFXIWindows[currentindex];
            ForceForegroundWindow(item.hwnd);
         } else {
            this.ActiveNextWindow();
         }
      }

      /// <summary>
      /// Toggles the current game window between the normal and maximize states
      /// </summary>
      internal void ToggleZoom() {
#if DEBUG
         System.Diagnostics.Trace.WriteLine("ACTION: Toggle Zoom");
#endif
         if (this.FFXIWindows.Count > 0) {
            //Flip to the next window if the pointer is out of bounds due to a game window unloading
            if (currentindex >= this.FFXIWindows.Count)
               ActiveNextWindow();

            IntPtr hWnd = this.FFXIWindows[currentindex].hwnd;
            if (WinAPI.IsZoomed(hWnd) || WinAPI.IsIconic(hWnd)) {
               WinAPI.ShowWindow(hWnd, WinAPI.ShowWindowFlags.SW_RESTORE);
            } else {
               WinAPI.ShowWindow(hWnd, WinAPI.ShowWindowFlags.SW_MAXIMIZE);
            }
         }
      }

      internal void ToggleMagnifiers() {
#if DEBUG
         System.Diagnostics.Trace.WriteLine("ACTION: Toggle Magnifiers");
#endif
         if (MagSetManager.AnyOpen()) {
            MagSetManager.CloseAll();
         } else {
            MagSetManager.OpenAll();
            ActivateCurrentWindow();
         }
      }

      internal void CycleMagSet() {
#if DEBUG
         System.Diagnostics.Trace.WriteLine("ACTION: Cycle Mag Set");
#endif
         MagSetManager.ActivateNextSet();
      }
      
      public void ExecuteCommand(ActionCommand Command) {
         switch(Command) {
            case ActionCommand.ActivateNext:
               this.ActiveNextWindow();
               break;
            case ActionCommand.ActivatePrev:
               this.ActivePreviousWindow();
               break;
            case ActionCommand.ToggleZoom:
               this.ToggleZoom();
               break;
            case ActionCommand.ToggleMagnifiers:
               this.ToggleMagnifiers();
               break;
            case ActionCommand.CycleSets:
               this.CycleMagSet();
               break;
            case ActionCommand.OpenMenu:
               fMenu menuwindow = new fMenu(this, this.joystick);
               menuwindow.Show();
               break;
         }
      }

      //==========================================================================================
      // Events
      //==========================================================================================
      void fMonitor_OptionsUpdated() {
         bindlist.Clear();
         bindlist[config.GetValue<int>("buttonflipnext" , -1)] = ActionCommand.ActivateNext;
         bindlist[config.GetValue<int>("buttonflipprev" , -1)] = ActionCommand.ActivatePrev;
         bindlist[config.GetValue<int>("buttonmaxtoggle", -1)] = ActionCommand.ToggleZoom;
         bindlist[config.GetValue<int>("buttonmagtoggle", -1)] = ActionCommand.ToggleMagnifiers;
         bindlist[config.GetValue<int>("buttoncycleset" , -1)] = ActionCommand.CycleSets;
         bindlist[config.GetValue<int>("buttonopenmenu" ,  5)] = ActionCommand.OpenMenu;              

         m_gather_mode = config.GetValue<int>("gathermode", 2);
         tGather.Enabled = (m_gather_mode == 1);
         tGather.Interval = config.GetValue<int>("gatherinterval", 1000);
      }

      void joystick_ButtonPressDown(Joy.Joystick Joystick, int ID) {
         if (bindlist.ContainsKey(ID))
            ExecuteCommand(bindlist[ID]);
      }

      void MagSetManager_SetChanged(MagSet set) {
         if (set != null) {
            foreach (ToolStripItem menuitem in this.miChangeMagSet.DropDownItems){
               if (menuitem.Tag == set) {
                  menuitem.Image = global::FFXISwitchMon.Properties.Resources.RightArrow;
                  menuitem.ImageTransparentColor = System.Drawing.Color.Fuchsia;
               } else {
                  menuitem.Image = null;
               }
            }
            foreach (ToolStripItem menuitem in this.miRemoveMagSet.DropDownItems){
               if (((MagSet)menuitem.Tag).Name == set.Name) {
                  menuitem.Image = global::FFXISwitchMon.Properties.Resources.RightArrow;
                  menuitem.ImageTransparentColor = System.Drawing.Color.Fuchsia;
               } else {
                  menuitem.Image = null;
               }
            }
         }
      }

      void MagSetManager_SetRemoved(MagSet set) {
         foreach (ToolStripItem menuitem in this.miChangeMagSet.DropDownItems){
            if (menuitem.Tag == set) {
               menuitem.Click -= new EventHandler(miChangeMagSet_SubItem_Click);
               miChangeMagSet.DropDownItems.Remove(menuitem);
               break;
            }
         }
         foreach (ToolStripItem menuitem in this.miRemoveMagSet.DropDownItems){
            if (menuitem.Tag == set) {
               menuitem.Click -= new EventHandler(miRemoveMagSet_SubItem_Click);
               miRemoveMagSet.DropDownItems.Remove(menuitem);
               break;
            }
         }
      }

      void MagSetManager_SetCreated(MagSet set) {
         ToolStripItem newitem;
         newitem = this.miChangeMagSet.DropDownItems.Add(set.Name);
         newitem.Click += new EventHandler(miChangeMagSet_SubItem_Click);
         newitem.Tag = set;

         newitem = this.miRemoveMagSet.DropDownItems.Add(set.Name);
         newitem.Click += new EventHandler(miRemoveMagSet_SubItem_Click);
         newitem.Tag = set;
      }

      private void miOptions_Click(object sender, EventArgs e) {
         if (opt == null || opt.IsDisposed || (!opt.IsDisposed && !opt.Created)) {
            opt = new fOptions(this);
            if (opt.ShowDialog(this) == DialogResult.OK && this.OptionsUpdated != null)
               this.OptionsUpdated();
         }
      }

      private void miRefresh_Click(object sender, EventArgs e) {
         EnumFFXIWindows();
			//Initialize Joysticks
			if (joys.Count > 0)
			{
				//Load the previous joystick
				if (config["lastjoy"] != null)
				{
					Guid joyguid = new Guid((string)config["lastjoy"]);
					joystick = joys.FindJoystick(joyguid);
				}
				//If no previous joystick the select the first one
				if (joystick == null)
				{
					joystick = joys[0];
					config["lastjoy"] = joystick.ID;
					config.Save();
				}
				//Begin polling joystick
				joystick.ButtonPressDown += new Joy.ButtonPressCallback(joystick_ButtonPressDown);
				joystick.Acquire();
			}
			else
			{
				joystick = null;
			}
		}

      private void miExit_Click(object sender, EventArgs e) {
         //Save state info
         MagSetManager.Save();
         //Save program config
         config.Save();

         //Deallocate objects
         MagSetManager.SetCreated -= new MagnifierSetCallback(MagSetManager_SetCreated);
         MagSetManager.SetRemoved -= new MagnifierSetCallback(MagSetManager_SetRemoved);
         MagSetManager.SetChanged -= new MagnifierSetCallback(MagSetManager_SetChanged);
         tStatus.Dispose();

         //Kill the thread
         Application.Exit();
      }

      private void miMagnifier_Click(object sender, EventArgs e) {
         MagSetManager.OpenNext();
      }
      
      private void miMagnifierOpenAll_Click(object sender, EventArgs e) {
         this.MagSetManager.OpenAll();
      }

      private void miMagnifierCloseAll_Click(object sender, EventArgs e) {
         this.MagSetManager.CloseAll();
      }

      private void tStatus_MouseDoubleClick(object sender, MouseEventArgs e) {
         if (e.Button == MouseButtons.Left) {
            miOptions_Click(sender, e);
         }
      }

      void tGather_Tick(object sender, EventArgs e) {
         EnumFFXIWindows();
      }

      private void miSaveNow_Click(object sender, EventArgs e) {
         MagSetManager.Save();
         config.Save();
         MessageBox.Show("Configuration has been successfully written.", "Switch Monitor", MessageBoxButtons.OK, MessageBoxIcon.Information);
      }

      private void miNewMagSet_Click(object sender, EventArgs e) {
         string newname;
         if (fInputBox.GetTextResult(out newname) == DialogResult.OK) {
            MagSet set = MagSetManager.CreateSet(newname);
            MagSetManager.ChangeSet(set);
         }
      }

      private void miChangeMagSet_SubItem_Click(object sender, EventArgs e) {
         try {
            ToolStripItem menuitem = (ToolStripItem)sender;
            MagSet set = (MagSet)menuitem.Tag;
            MagSetManager.ChangeSet(set);
         } catch (Exception ex) {
#if DEBUG
            System.Diagnostics.Trace.WriteLine("Error changing set: " + ex.Message);
#endif
         }
      }

      private void miRemoveMagSet_SubItem_Click(object sender, EventArgs e) {
         try {
            if (MagSetManager.SetCount > 1) {
               ToolStripItem menuitem = (ToolStripItem)sender;
               MagSet set = (MagSet)menuitem.Tag;

               if (MessageBox.Show("Do you really wish to permanently remove the set '" + set.Name + "'?", "Switch Monitor", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == DialogResult.Yes) {
                  MagSetManager.RemoveSet(set);
               }
            } else {
               MessageBox.Show("There is only one set available and cannot be removed. Create a new set first if you wish to remove it.", "Switch Monitor", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
         } catch (Exception ex) {
#if DEBUG
            System.Diagnostics.Trace.WriteLine("Error removing set: " + ex.Message);
#endif
         }
      }
   }
}